@@ -18,7 +18,7 @@ from yt_dlp.postprocessor.modify_chapters import ModifyChaptersPP
 
 from spotdl.types import Song
 from spotdl.utils.ffmpeg import FFmpegError, convert, get_ffmpeg_path
-from spotdl.utils.metadata import embed_metadata, MetadataError, get_song_metadata
+from spotdl.utils.metadata import embed_metadata, MetadataError
 from spotdl.utils.formatter import create_file_name, restrict_filename
 from spotdl.providers.audio.base import AudioProvider
 from spotdl.providers.lyrics import Genius, MusixMatch, AzLyrics
@@ -85,6 +85,8 @@ class Downloader:
         print_errors: bool = False,
         sponsor_block: bool = False,
         loop: Optional[asyncio.AbstractEventLoop] = None,
+        playlist_numbering: bool = False,
+        preserve_original_audio: bool = False,
     ):
         """
         Initialize the Downloader class.
@@ -93,8 +95,7 @@ class Downloader:
         - audio_provider: Audio providers to use.
         - lyrics_provider: The lyrics providers to use.
         - ffmpeg: The ffmpeg executable to use.
-        - variable_bitrate: The variable bitrate to use.
-        - constant_bitrate: The constant bitrate to use.
+        - bitrate: The bitrate to use.
         - ffmpeg_args: The ffmpeg arguments to use.
         - output_format: The output format to use.
         - threads: The number of threads to use.
@@ -110,6 +111,8 @@ class Downloader:
         - restrict: Whether to restrict the filename to ASCII characters.
         - print_errors: Whether to print errors on exit.
         - sponsor_block: Whether to remove sponsor segments using sponsor block postprocessor.
+        - playlist_numbering: Whether to convert tracks in a playlist into an album
+        - preserve_original_audio: Whether to preserve the original audio file
 
         ### Notes
         - `search-query` uses the same format as `output`.
@@ -188,6 +191,8 @@ class Downloader:
         self.sponsor_block = sponsor_block
         self.audio_providers_classes = audio_providers_classes
         self.progress_handler = ProgressHandler(NAME_TO_LEVEL[log_level], simple_tui)
+        self.playlist_numbering = playlist_numbering
+        self.preserve_original_audio = preserve_original_audio
 
         self.lyrics_providers: List[LyricsProvider] = []
         for lyrics_provider_class in lyrics_providers_classes:
@@ -345,7 +350,7 @@ class Downloader:
         # If it's None extract the current metadata
         # And reinitialize the song object
         if song.name is None and song.url:
-            song = reinit_song(song)
+            song = reinit_song(song, self.playlist_numbering)
 
         # Find song lyrics and add them to the song object
         lyrics = self.search_lyrics(song)
@@ -358,6 +363,9 @@ class Downloader:
         else:
             song.lyrics = lyrics
 
+        # Initalize the progress tracker
+        display_progress_tracker = self.progress_handler.get_new_tracker(song)
+
         # Create the output file path
         output_file = create_file_name(song, self.output, self.output_format)
         temp_folder = get_temp_path()
@@ -370,42 +378,34 @@ class Downloader:
         # we can skip the download
         if output_file.exists() and self.overwrite == "skip":
             self.progress_handler.log(f"Skipping {song.display_name}")
-            self.progress_handler.overall_completed_tasks += 1
-            self.progress_handler.update_overall()
+            display_progress_tracker.notify_download_skip()
             return song, None
 
         if output_file.exists() and self.overwrite == "metadata":
-            song_meta = get_song_metadata(output_file)
-            if song_meta is None:
-                self.progress_handler.debug(
-                    f"Metadata not found for {song.display_name}, " "overwriting file"
-                )
-            else:
-                self.progress_handler.debug(
-                    f"Metadata found for {song.display_name}, " "overwriting file"
-                )
-
             embed_metadata(
                 output_file=output_file, song=song, file_format=self.output_format
             )
 
+            self.progress_handler.log(f"Updated metadata for {song.display_name}")
+            display_progress_tracker.notify_complete()
+
             return song, output_file
 
         # Don't skip if the file exists and overwrite is set to force
         if output_file.exists() and self.overwrite == "force":
             self.progress_handler.debug(f"Overwriting {song.display_name}")
 
-        # Initalize the progress tracker
-        display_progress_tracker = self.progress_handler.get_new_tracker(song)
-
         # Create the output directory if it doesn't exist
         output_file.parent.mkdir(parents=True, exist_ok=True)
 
         try:
             if song.download_url is None:
-                url, audio_provider = self.search(song)
+                download_url, audio_provider = self.search(song)
             else:
-                url = song.download_url
+                # If the song object already has a download url
+                # we can skip the search, and just reinitialize the base
+                # audio provider to download the song
+                download_url = song.download_url
                 audio_provider = AudioProvider(
                     output_format=self.output_format,
                     cookie_file=self.cookie_file,
@@ -414,7 +414,7 @@ class Downloader:
                 )
 
             self.progress_handler.debug(
-                f"Downloading {song.display_name} using {url}, "
+                f"Downloading {song.display_name} using {download_url}, "
                 f"audio provider: {audio_provider.name}"
             )
 
@@ -424,14 +424,17 @@ class Downloader:
             )
 
             # Download the song using yt-dlp
-            download_info = audio_provider.get_download_metadata(url, download=True)
+            download_info = audio_provider.get_download_metadata(
+                download_url, download=True
+            )
+
             temp_file = Path(
                 temp_folder / f"{download_info['id']}.{download_info['ext']}"
             )
 
             if download_info is None:
                 self.progress_handler.debug(
-                    f"No download info found for {song.display_name}, url: {url}"
+                    f"No download info found for {song.display_name}, url: {download_url}"
                 )
 
                 raise LookupError(
@@ -440,19 +443,37 @@ class Downloader:
 
             display_progress_tracker.notify_download_complete()
 
+            bitrate: Optional[str] = (
+                self.bitrate if self.bitrate else f"{int(download_info['abr'])}k"
+            )
+
+            # Ignore the bitrate if the preserve original audio
+            # option is set to true
+            if self.preserve_original_audio:
+                bitrate = None
+
             success, result = convert(
-                temp_file,
-                output_file,
-                self.ffmpeg,
-                self.output_format,
-                self.bitrate,
-                self.ffmpeg_args,
-                display_progress_tracker.ffmpeg_progress_hook,
+                input_file=temp_file,
+                output_file=output_file,
+                ffmpeg=self.ffmpeg,
+                output_format=self.output_format,
+                bitrate=bitrate,
+                ffmpeg_args=self.ffmpeg_args,
+                progress_handler=display_progress_tracker.ffmpeg_progress_hook,
             )
 
             # Remove the temp file
             if temp_file.exists():
-                temp_file.unlink()
+                try:
+                    temp_file.unlink()
+                except (PermissionError, OSError) as exc:
+                    self.progress_handler.debug(
+                        f"Could not remove temp file: {temp_file}, error: {exc}"
+                    )
+
+                    raise DownloaderError(
+                        f"Could not remove temp file: {temp_file}, possible duplicate song"
+                    ) from exc
 
             if not success and result:
                 # If the conversion failed and there is an error message
@@ -484,29 +505,38 @@ class Downloader:
 
             # Set the song's download url
             if song.download_url is None:
-                song.download_url = download_info["webpage_url"]
+                song.download_url = download_url
 
             display_progress_tracker.notify_conversion_complete()
 
+            # SponsorBlock post processor
             if self.sponsor_block:
+                # Initialize the sponsorblock post processor
                 post_processor = SponsorBlockPP(
                     audio_provider.audio_handler, SPONSOR_BLOCK_CATEGORIES
                 )
 
+                # Run the post processor to get the sponsor segments
                 _, download_info = post_processor.run(download_info)
                 chapters = download_info["sponsorblock_chapters"]
+
+                # If there are sponsor segments, remove them
                 if len(chapters) > 0:
                     self.progress_handler.log(
                         f"Removing {len(chapters)} sponsor segments for {song.display_name}"
                     )
 
+                    # Initialize the modify chapters post processor
                     modify_chapters = ModifyChaptersPP(
                         audio_provider.audio_handler,
                         remove_sponsor_segments=SPONSOR_BLOCK_CATEGORIES,
                     )
 
+                    # Run the post processor to remove the sponsor segments
+                    # this returns a list of files to delete
                     files_to_delete, download_info = modify_chapters.run(download_info)
 
+                    # Delete the files that were created by the post processor
                     for file_to_delete in files_to_delete:
                         Path(file_to_delete).unlink()
 
@@ -524,8 +554,18 @@ class Downloader:
             )
 
             return song, output_file
-        except Exception as exception:
-            display_progress_tracker.notify_error(traceback.format_exc(), exception)
+        except (Exception, UnicodeEncodeError) as exception:
+            if isinstance(exception, UnicodeEncodeError):
+                exception_cause = exception
+                exception = DownloaderError(
+                    "You may need to add PYTHONIOENCODING=utf-8 to your environment"
+                )
+
+                exception.__cause__ = exception_cause
+
+            display_progress_tracker.notify_error(
+                traceback.format_exc(), exception, True
+            )
             self.errors.append(
                 f"{song.url} - {exception.__class__.__name__}: {exception}"
             )
